/*
** Low cost Z86E04/E08 programmer.
** Modified from David Tait(david.tait@man.ac.uk)'s pp.c
** 
** Features:
** 1. Dump E04/E08 program memory content to screen or to a binary file.
** 2. Program E04/E08 with data from a binary/hex file.
** 3. Prefill program memory with %00 or %FF.
** 4. Program Low EMI Mode and/or Rom Protect Mode.
**
** Future Modifications:
** 1. reduce the transistor counts by using a inverting open-collect driver
**    (7405) - by replacing the 7407 by 7405, 2/3 transistors can be eliminated
** 2. save read data to Intel-hex file.
**
** Disclaimer:
** The program/schematics are provide "as is" without warranty of any kind,
** either expressed or implied.
*/


#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <io.h>
#ifdef GCC
#include <pc.h>
#include <go32.h>
#endif

#define outp	outportb
#define inp	inportb

#define PSIZE   	4096
#define PROG_LOWEMI	0
#define PROG_ROMPROT	1

#define IN      	8
#define VPP     	8
#define VDD     	4
#define CLK     	2
#define OUT     	1
#define SDataOut	4
#define SClk		8

#define vppon           (d_bits &= ~VPP)
#define vppoff          (d_bits |= VPP)
#define vddon           (d_bits &= ~VDD)
#define vddoff          (d_bits |= VDD)
#define clkhi           (d_bits &= ~CLK)
#define clklo           (d_bits |= CLK)
#define outhi           (d_bits &= ~OUT)
#define outlo           (d_bits |= OUT)

#define SDataHi		(ControlReg |=  SDataOut)
#define SDataLo		(ControlReg &= ~SDataOut)
#define SClkLo		(ControlReg &= ~SClk)
#define SClkHi		(ControlReg |=  SClk)
#define Load		(ControlReg &= ~0x1)
#define Shift		(ControlReg |=  0x1)
#define E08OEHi		(DataReg |=  0x1)
#define E08OELo		(DataReg &= ~0x1)
#define E08CEHi		(DataReg &= ~0x20)
#define E08CELo		(DataReg |=  0x20)
#define E08EPMVHi	(DataReg |=  0x80); (ControlReg &= ~0x2)
#define E08EPMHi	(DataReg &= ~0x80); (ControlReg &= ~0x2)
#define E08EPMLo	(DataReg |=  0x80); (ControlReg |=  0x2)
#define E08ClkHi	(DataReg |=  0x4)
#define E08ClkLo	(DataReg &= ~0x4)
#define E08ClrHi	(DataReg |=  0x8)
#define E08ClrLo	(DataReg &= ~0x8)
#define E08Vcc6V	(DataReg &= ~0x10)
#define E08Vcc5V	(DataReg |=  0x10)
#define E08VppHi	(DataReg &= ~0x40)
#define E08VppOD	(DataReg |=  0x40)
#define E08PGMHi	(DataReg |=  0x2)
#define E08PGMLo	(DataReg &= ~0x2)

#define AssertDReg	(outp(d_reg,DataReg))
#define AssertCReg	(outp(c_reg,ControlReg))
#define ReadSReg        ((inp(s_reg) & 0x8))


#define LowEMI	 	0x0001
#define ReadMode 	0x0002
#define ProtectMode	0x0004
#define NoBlankCheck	0x0008
#define BinFile		0x0010
#define PreFill00	0x0020
#define OptionMask	(LowEMI+ReadMode+ProtectMode+NoBlankCheck+BinFile+PreFill00)
int MainFlag = 0;
int EndAddr = 0x07ff;

int progbuf[PSIZE];
int databuf[PSIZE];

int d_bits = 0;
int d_reg;
int s_reg;
int c_reg;
int ControlReg=0;
int DataReg=0;		     
int check;
int iport = 0;

char *progname = "Zilog Z86E04/E08 Programmer";
char *version = "Version 0.0";
char *copyright = "Copyright (C) 1995 Anonymous";
char *email = "winter@asiaonline.net";

void tiny_delay(int ticks)
{
   while ( ticks-- );
}
			 
void idle_mode(void)
{
   E08OEHi; E08CELo; E08Vcc5V; E08VppOD; E08EPMLo; E08PGMHi;
   E08ClkLo; E08ClrLo; Load;
   SDataLo; SClkLo;
   AssertDReg; AssertCReg;
}


void quit(char *s)
{
   fprintf(stderr,"Error: %s\n",s);
   idle_mode();
   exit(1);
}
      
void usage(void)
{
   printf("%s  %s  %s\n\n",progname,version,copyright);
   printf("Usage: pe  [-48lpfrb0]  [file1] \n\n");
   printf("    file1 = hex or binary data file\n\n");
   printf("        4 - target is E04: address %%000-%%03FF\n");
   printf("        8 - target is E08: address %%000-%%07FF\n");
   printf("        l - enable low EMI mode\n");
   printf("        p - enable rom protect mode\n");
   printf("        f - force mode, skip initial blank check\n");
   printf("        r - read OTP data and display\n");
   printf("        b - binary file\n");
   printf("        0 - prefile memory with %%00, else prefill with %%FF\n");
   printf("  Defaults: -8\n");
   printf("  Report Bugs to %s\n\n",email);
   printf("  Examples:\n");
   printf("  pe -4r                read E04 and dump data to screen\n");
   printf("  pe -4r e04data.bin    read E04 and write data to file e04data.bin\n");
   printf("  pe e08data.hex        blank check and then write E08 with data from\n");
   printf("                        hex file e08data.hex\n");
   printf("  pe -4fb e04data.bin   write E04 with binary data from e04data.bin\n");
   idle_mode();
   exit(1);
}

void setup(void)
{
   int i, b;
#ifdef GCC
   dosmemget(0x408+iport, 2, &d_reg);
/*   d_reg = 0x378; */
#else
   d_reg = *(unsigned int far *)(0x00400008+iport);  /* find the base address of LPT port */
#endif
   
   s_reg = d_reg+1;
   c_reg = s_reg+1;

   switch ( d_reg ) {  /* I don't want to blow anything up, so check port */
       case 0x3BC:
       case 0x378:
       case 0x278: break;
       default:  fprintf(stderr,"pp: LPT at 0x%04X\n",d_reg);
		 quit("Unlikely LPT address");
   }

/*   printf("LPT port at %4X, %4X, %4X\n", d_reg, s_reg, c_reg); */
   idle_mode();
}


void get_option(char *s)
{
   char *sp;
   for ( sp=s; *sp; ++sp )
      switch ( *sp ) {
	 case 'l':
	 case 'L': MainFlag |= LowEMI; break;
	 case 'r':
	 case 'R': MainFlag |= ReadMode; break;
	 case '4': EndAddr = 0x03ff; break;
	 case '8': EndAddr = 0x07ff; break;
         case 'p':
       	 case 'P': MainFlag |= ProtectMode; break;
         case 'f':
         case 'F': MainFlag |= NoBlankCheck; break;
         case 'b':
         case 'B': MainFlag |= BinFile; break;
         case '0': MainFlag |= PreFill00; break;
         case '-':
	 case '/': break;
	 default: usage();
      }
}



int hexdigit(FILE *fp)
{
   int c;

   if ( (c=getc(fp)) == EOF ) quit("Unexpected EOF");

   c -= c>'9'? 'A'-10: '0';
   if ( c<0 || c>0xF ) quit("Hex digit expected");
   return c;
}

int hexbyte(FILE *fp)
{
   int b;

   b = hexdigit(fp);
   b = (b<<4) + hexdigit(fp);
   check += b;               /* nasty side-effect, but useful trick */
   return b;
}

unsigned hexword(fp)
FILE *fp;
{
   unsigned w;

   w = hexbyte(fp);
   w = (w<<8) + hexbyte(fp);
   return w;
}

void LoadBin(FILE *fp, int buf[], int bufsize)
{
   int c, i;

   for (i=0; ((c=getc(fp))!=EOF) && (i < bufsize); i++) buf[i] = c;
}

void loadhex(FILE *fp, int buf[], int bufsize)
{
   int type, nbyte, i;
   unsigned address;

   type = 0;
   while ( type != 1 ) {
      if ( getc(fp) != ':' ) quit("Expected ':'");
      check = 0;
      nbyte = hexbyte(fp);		/* no of data byte */
      address = hexword(fp);		/* starting address */
      type = hexbyte(fp);		/* operation type */
      for ( i=0; i<nbyte; ++i, ++address ) {
	 if ( address >= bufsize ) quit("Data File Too Large");
	 buf[address] = hexbyte(fp);
      }
      (void) hexbyte(fp);	/* check sum */
      (void) getc(fp);		/* LF */
      if (check&0xFF) quit("Checksum error");
   }
}

void DumpBuf(int buf[])
{
   int i;

   printf("0000 - ");
   for (i=0; i<=EndAddr; i++)  {
      printf("%02X ", buf[i]);
      if (((i&15)==15) && (i!=EndAddr)) printf("\n%04X - ", i+1);
   }
   putchar('\n');
}

void PreFill(int buf[], int v)
{
   int i;
   for (i=0; i<=EndAddr; i++) buf[i] = v;
}

void ClrAddrCnt(void)
{
   E08ClrHi; AssertDReg;
   E08ClkHi; AssertDReg;
   tiny_delay(10);
   E08ClkLo; AssertDReg;
   E08ClrLo; AssertDReg;
   tiny_delay(10);
}

int RxE08(void)
{
   int i, SerBuf;
   
   E08OELo; AssertDReg;	tiny_delay(10);		/* 3-state */
   Load; SClkLo; AssertCReg; tiny_delay(100);
   SClkHi; AssertCReg; tiny_delay(10);
   SClkLo; AssertCReg; tiny_delay(10);
   Shift; AssertCReg; tiny_delay(10);
   E08OEHi; AssertDReg; tiny_delay(10);
   for (i=0; i<=7; i++)  {
/*       printf("read %04X %04X\n", SerBuf, ReadSReg);*/
      SerBuf = (SerBuf<<1) | (((ReadSReg)==0) ? 0 : 1);
      SClkHi; AssertCReg; tiny_delay(10);
      SClkLo; AssertCReg; tiny_delay(10);
   }
   return (SerBuf & 0xff);
}

int ReadOTP(void)
{
   int i, RetFlag = 0, CS = 0;
   
   E08PGMHi; AssertDReg;
   tiny_delay(10);
   E08VppHi; E08EPMVHi; E08Vcc6V; AssertDReg, AssertCReg;
   ClrAddrCnt();
   E08CELo; AssertDReg;

   for (i=0; i<=EndAddr; i++) {
      databuf[i] =  RxE08();
      if (databuf[i] != 0xff) RetFlag = 1;
      CS += databuf[i];
      E08ClkHi; AssertDReg; tiny_delay(10);
      E08ClkLo; AssertDReg; tiny_delay(10);
   }
   if (!RetFlag) return ((CS & 0xffff) | 0x10000);
   else return (CS & 0xffff);
   idle_mode();
}

void TxSerData(int SerBuf)
{
   int i;

   Shift;
   SDataLo;
   SClkLo;
   AssertCReg;
/*    printf ("%4X\n", SerBuf); */
   for (i=1; i<=8; i++)  {
      if (SerBuf & 0x80) {
	 SDataHi;		/* set data */
/* 	 putchar('1'); */
      } else {
	 SDataLo;
/* 	 putchar('0'); */
      }
      AssertCReg; tiny_delay(50);
      SClkHi; AssertCReg; tiny_delay(50);
      SClkLo; AssertCReg; tiny_delay(50);
      SerBuf <<= 1;
   }
   SDataLo;
   AssertCReg;
}

void WriteAByte(int data, int PulseWidth)
{
      TxSerData(data);
      E08PGMLo; AssertDReg; delay(PulseWidth);
      E08PGMHi; AssertDReg; tiny_delay(10);
}
      

void ProgOTP(void)
{
   int i, CS, WrState, trial;
   E08VppHi; E08EPMVHi; E08Vcc6V; E08CELo; E08OEHi; E08PGMHi;
   AssertDReg; AssertCReg;
   ClrAddrCnt();
   tiny_delay(10);
   printf("Writing %%0000");
   for (i=0; i<=EndAddr; i++) {
      printf("\rWriting %%%04X", i); fflush(stdout);
      for (trial=1, WrState=0; (trial<=25 && WrState==0); trial++) {
	 WriteAByte(progbuf[i], 1);	/* 1ms write pulse */
/* 	 printf("RxE08: %X\n", RxE08()); */
	 if (RxE08() == progbuf[i]) WrState=1;
      }
      if (WrState==0) {
	 fprintf(stderr, "\n > 25 write error\n");
/* 	 printf ("RxE08 progbuf[%d] %X %X\n", i, RxE08(), progbuf[i]); */
	 idle_mode();
	 exit(1);
      }
      else {
	 WriteAByte(progbuf[i], 3);	/* 3ms write pulse */
         if (RxE08() != progbuf[i]) {
/*	    printf("RxE08 progbuf[%d] %X %X\n", i, RxE08(), progbuf[i]); */
	    quit("3x Pulse Write Error");
	 }
      }
      E08ClkHi; AssertDReg; tiny_delay(10);	/* inc E08 address counter */
      E08ClkLo; AssertDReg; tiny_delay(10);
   }
   printf("\rVerifing.....\n");
   E08Vcc5V; AssertDReg; tiny_delay(10);
   CS = ReadOTP() & 0xffff;
   for (i=0; i<=EndAddr; i++)
      if (progbuf[i] != databuf[i])  {
	 printf("Verify Error at %04X\n", i);
	 quit("");
      }
   printf("Verify OK! CheckSum is %04X\n", CS);
   idle_mode();
}

void ProgEMIProt(int ProgMode)
{
   if (ProgMode==PROG_LOWEMI) {
      printf("Low EMI Mode - Done !\n");
      E08Vcc6V; E08VppHi; E08CEHi; E08OEHi; E08EPMVHi; E08PGMHi;
   } else {
      printf("Rom Protect Mode - Done !\n");
      E08Vcc6V; E08VppHi; E08CEHi; E08OEHi; E08EPMHi; E08PGMHi;
   }
   AssertDReg; AssertCReg;
   delay(1);
   E08PGMLo; AssertDReg;	       /* 78ms program pulses */
   delay(160);
   E08PGMHi; AssertDReg;
   idle_mode();
}

void main(int argc, char *argv[])
{
   FILE *objfp, *fpout;
   int i, c;
   char DataFile[80], FileMode[8];
   
   DataFile[0] = '\0';		/* init empty file name */
   for ( i=1; i<argc; ++i ) {
      if ((c = *argv[i]) == '-' || c == '/')
	 get_option(++argv[i]);
      else
      	strcpy (DataFile, argv[i]);
   }
   
   
   if (MainFlag != ReadMode)  {
      if (MainFlag & BinFile) strcpy (FileMode, "rb");
      else strcpy (FileMode, "r");
      if ((objfp = fopen(DataFile, FileMode)) == NULL )
	    quit("Can't open Data File");
   }

   setup();

   PreFill(progbuf, (MainFlag & PreFill00) ? 0x00 : 0xff);
   
   if (!(MainFlag & ReadMode))  {
      if (MainFlag & BinFile) LoadBin(objfp, progbuf, PSIZE);
      else loadhex(objfp, progbuf, PSIZE);
   }

/*   DumpBuf(progbuf);	print data */
  
   printf("%s  %s  %s\n",progname,version,copyright);

   printf("Insert %s and press a key to start (^C to abort) ...\n",
	  (EndAddr==0x03ff) ? "E04" : "E08");
   if (getch()==0x3) quit("Abort!!!");
   printf("\nPlease Wait...\n");
   
/* 
 * Read OTP function
 */ 
   if (MainFlag & ReadMode)  {
      printf ("CheckSum is %04X\n", ReadOTP() & 0xffff);
      if (DataFile[0] != '\0')  {	/* write binary file */
	 if ((fpout=fopen(DataFile, "wb"))==NULL) quit("File Creation Error");
	 for (i=0; i<=EndAddr; i++) putc(databuf[i], fpout);
	 fclose(fpout);
      } else DumpBuf(databuf);
      exit(0);
   } 
   
/*
 * Program OTP function
 */
   if (!(MainFlag & NoBlankCheck)) 
     if (ReadOTP() & 0x10000)
     quit("Blank Check Error. Use -f to force program");
   ProgOTP();
/*
 * Successfully programmed a OTP. See if Low EMI/Rom Protect mode is required
 */
 if (MainFlag & LowEMI) ProgEMIProt(PROG_LOWEMI);
 if (MainFlag & ProtectMode) ProgEMIProt(PROG_ROMPROT);
}
